from .contract import Contract
from .multicall import Multicall
from .disperse import Disperse
from .erc20 import ERC20
from .erc721 import ERC721

__all__ = [
    "Contract",
    "Multicall",
    "Disperse",
    "ERC20",
    "ERC721",
]



from pathlib import Path
from typing import Iterable
from hexbytes import HexBytes

import eth_abi
from eth_account.messages import encode_defunct
from eth_account.account import LocalAccount
from eth_utils import to_checksum_address
from eth_typing import AnyAddress, ChecksumAddress
from web3.types import BlockParams, BlockIdentifier, HexStr

from .file import load_lines

BLOCK_PARAMS = ("latest", "earliest", "pending", "safe", "finalized")


def sign_message(message: str, account: LocalAccount) -> str:
    message = encode_defunct(text=message)
    signed_message = account.sign_message(message)
    return signed_message.signature.hex()


def decode_string_or_bytes32(data: bytes) -> str:
    try:
        return eth_abi.decode(["string"], data)[0]
    except OverflowError:
        name = eth_abi.decode(["bytes32"], data)[0]
        end_position = name.find(b"\x00")
        if end_position == -1:
            return name.decode()
        else:
            return name[:end_position].decode()


def to_checksum_addresses(addresses: Iterable[AnyAddress or str]) -> list[ChecksumAddress]:
    return [to_checksum_address(address) for address in addresses]


def addresses_from_file(filepath: Path | str) -> list["ChecksumAddress"]:
    return to_checksum_addresses([address.strip() for address in load_lines(filepath)])


GAS_CALL_DATA_ZERO_BYTE = 4
GAS_CALL_DATA_BYTE = 16  # 68 before Istanbul


def estimate_data_gas(data: bytes):
    if isinstance(data, str):
        data = HexBytes(data)

    gas = 0
    for byte in data:
        if not byte:
            gas += GAS_CALL_DATA_ZERO_BYTE
        else:
            gas += GAS_CALL_DATA_BYTE
    return gas


def hex_block_identifier(block_identifier: BlockIdentifier) -> HexStr | BlockParams:
    if block_identifier in BLOCK_PARAMS:
        return block_identifier
    elif isinstance(block_identifier, int):
        block_identifier = hex(block_identifier)
    elif isinstance(block_identifier, bytes):
        block_identifier = HexBytes(block_identifier).hex()
    return HexStr(block_identifier)



from . import contract, utils
from .chain import Chain
from .wallet import Wallet
from .proxy import Proxy
# Exceptions
from .contract.multicall import MulticallFailed
from .batch_call import JSONRPCException


__all__ = [
    "contract",
    "utils",
    "Chain",
    "Wallet",
    "Proxy",
    "MulticallFailed",
    "JSONRPCException",
]



import asyncio
from typing import Iterable, Any, AsyncIterable
from typing import TYPE_CHECKING

import aiohttp
import eth_abi
from eth_abi.exceptions import DecodingError
from eth_typing import ChecksumAddress, HexStr, Address, Hash32
from hexbytes import HexBytes
from web3._utils.abi import map_abi_data
from web3._utils.method_formatters import (
    block_formatter,
    receipt_formatter,
    transaction_result_formatter,
)
from web3._utils.normalizers import BASE_RETURN_NORMALIZERS
from web3.contract.async_contract import AsyncContractFunction
from web3.types import (
    BlockData,
    BlockIdentifier,
    TxData,
    TxParams,
    TxReceipt,
    Wei,
)

from .utils import chunks, hex_block_identifier

if TYPE_CHECKING:
    from .chain import Chain


def build_payload(index: int, method: str, params: list) -> dict:
    return {"id": index, "jsonrpc": "2.0", "method": method, "params": params}


def process_results(results, *, raise_exceptions: bool = True) -> Iterable:
    """Process result from the batch request."""
    if isinstance(results, dict) and "error" in results:
        raise JSONRPCException(results["error"]["code"], results["error"]["message"])

    # Nodes like Erigon send back results out of order
    for result in sorted(results, key=lambda x: x["id"]):
        if "error" in result:
            e = JSONRPCException(result["error"]["code"], result["error"]["message"])
            if raise_exceptions:
                raise e
            else:
                yield e
        else:
            yield result["result"]


def decode_and_normalize_eth_call_result(result, output_type):
    """Decodes and normalizes a result based on the output_type."""
    decoded_values = eth_abi.decode(output_type, HexBytes(result))
    normalized_data = map_abi_data(BASE_RETURN_NORMALIZERS, output_type, decoded_values)
    return normalized_data[0] if len(normalized_data) == 1 else normalized_data


def process_eth_call_results(
        results,
        output_types,
        *,
        raise_exceptions: bool = True,
) -> Iterable:
    for result, output_type in zip(results, output_types):
        try:
            yield decode_and_normalize_eth_call_result(result, output_type)
        except (DecodingError, OverflowError):
            e = JSONRPCException(result["error"]["code"], f'DecodingError, cannot decode')
            if raise_exceptions:
                raise e
            else:
                yield e


class JSONRPCException(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        super().__init__(f'(code {self.code}) {self.message}')


class Manager:
    def __init__(self, chain: "Chain"):
        self.chain = chain
        self.rpc = chain.rpc
        self.w3 = chain.w3


class BatchCallManager(Manager):

    def __init__(
            self,
            chain: "Chain",
            batch_request_size: int = 500,
            batch_request_delay: int = 1,
    ):
        super().__init__(chain)
        self.batch_request_size = batch_request_size
        self.batch_request_delay = batch_request_delay

    async def request(
            self,
            payloads: Iterable[dict],
            *,
            raise_exceptions: bool = True,
            batch_size: int = None,
            delay: int = None,
    ) -> AsyncIterable[dict[str: dict, str: Any]]:
        if not payloads:
            return

        batch_size = batch_size or self.batch_request_size
        delay = delay or self.batch_request_delay

        async with aiohttp.ClientSession() as session:
            for chunk in chunks(payloads, batch_size):
                response = await session.post(self.rpc, json=chunk)
                results = await response.json()
                for payload, response in zip(chunk, process_results(results, raise_exceptions=raise_exceptions)):
                    result = {"payload": payload}
                    if isinstance(response, JSONRPCException):
                        result["exception"] = response
                    else:
                        result["result"] = response
                    yield result
                await asyncio.sleep(delay)

    async def contract_request(
            self,
            contract_functions: Iterable[AsyncContractFunction],
            block_identifier: BlockIdentifier = "latest",
            from_: ChecksumAddress | str = None,
            *,
            raise_exceptions: bool = True,
            batch_size: int = None,
            delay: int = None,
    ) -> AsyncIterable[dict[str: AsyncContractFunction, str: Any]]:
        if not contract_functions:
            return

        block_identifier = hex_block_identifier(block_identifier)

        payloads = []
        output_types = []
        for i, contract_function in enumerate(contract_functions):
            if not contract_function.address:
                raise ValueError(
                    f"Missing address for batch_call in `{contract_function.fn_name}`: {contract_function}")

            data = (await contract_function.build_transaction({"gas": 0, "gasPrice": 0}))["data"]
            tx_params = {"to": contract_function.address, "data": data}
            if from_: tx_params["from"] = from_
            payloads.append(build_payload(i, "eth_call", [tx_params, block_identifier]))

            output_type = [output["type"] for output in contract_function.abi["outputs"]]
            output_types.append(output_type)

        results = [result async for result in self.request(
            payloads,
            raise_exceptions=raise_exceptions,
            batch_size=batch_size,
            delay=delay,
        )]
        results = process_eth_call_results(results, output_types, raise_exceptions=raise_exceptions)
        for contract_function, response in zip(contract_functions, results):
            result = {"contract_function": contract_function}
            if isinstance(response, JSONRPCException):
                result["exception"] = response
            else:
                result["result"] = response
            yield result

    ################################################################################
    # Batch request shortcuts
    ################################################################################

    async def balances(
            self,
            addresses: Iterable[ChecksumAddress | str],
            block_identifier: BlockIdentifier = "latest",
            **kwargs,
    ) -> AsyncIterable[dict[str: ChecksumAddress | str, str: Wei | JSONRPCException]]:
        if not addresses:
            return

        payloads = [build_payload(i, "eth_getBalance", [address, hex_block_identifier(block_identifier)])
                    for i, address in enumerate(addresses)]
        balances = [balance async for balance in self.request(payloads, **kwargs)]

        for address, response in zip(addresses, balances):
            balance_data = {"address": address}
            if "exception" in response:
                balance_data["exception"] = response["exception"]
            else:
                balance_data["balance"] = Wei(int(response["result"], 16))
            yield balance_data

    async def txs(
            self,
            tx_hashes: Iterable[Hash32 | HexBytes | HexStr],
            **kwargs,
    ) -> AsyncIterable[dict[str: Hash32 | HexBytes | HexStr, str: TxData | JSONRPCException]]:
        if not tx_hashes:
            return

        payloads = [build_payload(i, "eth_getTransactionByHash", [HexBytes(tx_hash).hex()])
                    for i, tx_hash in enumerate(tx_hashes)]
        txs = [tx async for tx in self.request(payloads, **kwargs)]

        for tx_hash, response in zip(tx_hashes, txs):
            tx_data = {"tx_hash": tx_hash}
            if "exception" in response:
                tx_data["exception"] = response["exception"]
            else:
                tx_data["tx"] = transaction_result_formatter(response["result"])
            yield tx_data

    async def tx_receipts(
            self,
            tx_hashes: Iterable[Hash32 | HexBytes | HexStr],
            **kwargs,
    ) -> AsyncIterable[dict[str: Hash32 | HexBytes | HexStr, str: TxReceipt | JSONRPCException]]:
        if not tx_hashes:
            return

        payloads = [build_payload(i, "eth_getTransactionReceipt", [HexBytes(tx_hash).hex()])
                    for i, tx_hash in enumerate(tx_hashes)]
        tx_receipts = [tx_receipt async for tx_receipt in self.request(payloads, **kwargs)]

        for tx_hash, response in zip(tx_hashes, tx_receipts):
            tx_receipt_data = {"tx_hash": tx_hash}
            if "exception" in response:
                tx_receipt_data["exception"] = response["exception"]
            else:
                tx_receipt_data["tx"] = receipt_formatter(response["result"])
            yield tx_receipt_data

    async def blocks(
            self,
            block_identifiers: Iterable[BlockIdentifier],
            full_transactions: bool = False,
            **kwargs,
    ) -> AsyncIterable[dict[str: BlockIdentifier, str: BlockData | None | JSONRPCException]]:
        if not block_identifiers:
            return

        payloads = []
        for i, block_identifier in enumerate(block_identifiers):
            is_int = isinstance(block_identifier, int)
            method = "eth_getBlockByNumber" if is_int else "eth_getBlockByHash"
            block_identifier = block_identifier if is_int else hex_block_identifier(block_identifier)
            payloads.append(build_payload(i, method, [block_identifier, full_transactions]))

        blocks_data = [block_data async for block_data in self.request(payloads, **kwargs)]

        for block_identifier, response in zip(block_identifiers, blocks_data):
            block_data = {"block_identifier": block_identifier}
            if "exception" in response:
                block_data["exception"] = response["exception"]
            else:
                if "extraData" in response["exception"]:
                    del block_data["extraData"]  # Remove extraData, raises some problems on parsing
                block_data["block"] = receipt_formatter(block_formatter(response["result"]))
            yield block_data



from ..utils.file import load_json
from ._paths import ABI_DIR


MULTICALL_V3_ABI = load_json(ABI_DIR / "multicall_v3.json")
DISPERSE_ABI     = load_json(ABI_DIR / "disperse.json")
ERC1155_ABI      = load_json(ABI_DIR / "erc1155.json")
ERC721_ABI       = load_json(ABI_DIR / "erc721.json")
ERC20_ABI        = load_json(ABI_DIR / "erc20.json")



"""More about MulticallV3: https://github.com/mds1/multicall
"""
from dataclasses import dataclass
from typing import Any, Sequence

import eth_abi
from eth_abi.exceptions import DecodingError
from eth_typing import BlockIdentifier, BlockNumber, ChecksumAddress
from hexbytes import HexBytes
from web3._utils.abi import map_abi_data
from web3._utils.normalizers import BASE_RETURN_NORMALIZERS
from web3.contract.async_contract import AsyncContract, AsyncContractFunction
from web3.exceptions import ContractLogicError

from ._abi import MULTICALL_V3_ABI
from .contract import Contract

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..chain import Chain


MULTICALL_V3_CONTRACT_ADDRESS = "0xcA11bde05977b3631167028862bE2a173976CA11"


@dataclass
class MulticallResult:
    success: bool
    return_data: bytes | None


@dataclass
class MulticallDecodedResult:
    success: bool
    return_data_decoded: Any | None


class MulticallFailed(Exception):
    pass


class Multicall(Contract):
    def __init__(
        self,
        chain: "Chain",
        address: ChecksumAddress | str = None,
        abi=None,
    ):
        address = address or MULTICALL_V3_CONTRACT_ADDRESS
        abi = abi or MULTICALL_V3_ABI
        super().__init__(chain, address, abi)

    @staticmethod
    def _build_payload(
        contract_functions: Sequence[AsyncContractFunction],
    ) -> tuple[list[tuple[ChecksumAddress, bytes]], list[list[Any]]]:
        targets_with_data = []
        output_types = []
        for contract_function in contract_functions:
            targets_with_data.append(
                (
                    contract_function.address,
                    HexBytes(contract_function._encode_transaction_data()),
                )
            )
            output_types.append(
                [output["type"] for output in contract_function.abi["outputs"]]
            )

        return targets_with_data, output_types

    @staticmethod
    def _decode_data(output_type: Sequence[str], data: bytes) -> Any | None:
        """

        :param output_type:
        :param data:
        :return:
        :raises: DecodingError
        """
        if data:
            try:
                decoded_values = eth_abi.decode(output_type, data)
                normalized_data = map_abi_data(
                    BASE_RETURN_NORMALIZERS, output_type, decoded_values
                )
                if len(normalized_data) == 1:
                    return normalized_data[0]
                else:
                    return normalized_data
            except DecodingError:
                print(
                    "Cannot decode %s using output-type %s", data, output_type
                )
                return data

    async def _aggregate(
        self,
        targets_with_data: Sequence[tuple[ChecksumAddress, bytes]],
        block_identifier: BlockIdentifier = "latest",
    ) -> tuple[BlockNumber, list[Any | None]]:
        """
        :param targets_with_data: List of target `addresses` and `data` to be called in each Contract
        """
        aggregate_parameter = [
            {"target": target, "callData": data} for target, data in targets_with_data
        ]
        try:
            return await self.contract.functions.aggregate(aggregate_parameter).call(
                block_identifier=block_identifier
            )
        except (ContractLogicError, OverflowError):
            raise MulticallFailed

    async def aggregate(
        self,
        contract_functions: Sequence[AsyncContractFunction],
        block_identifier: BlockIdentifier = "latest",
    ) -> tuple[BlockNumber, list[Any | None]]:
        targets_with_data, output_types = self._build_payload(contract_functions)
        block_number, results = await self._aggregate(
            targets_with_data, block_identifier=block_identifier
        )
        decoded_results = [
            self._decode_data(output_type, data)
            for output_type, data in zip(output_types, results)
        ]
        return block_number, decoded_results

    async def _try_aggregate(
        self,
        targets_with_data: Sequence[tuple[ChecksumAddress, bytes]],
        require_success: bool = False,
        block_identifier: BlockIdentifier = "latest",
    ) -> list[MulticallResult]:
        aggregate_parameter = [
            {"target": target, "callData": data} for target, data in targets_with_data
        ]

        try:
            result = await self.functions.tryAggregate(
                require_success, aggregate_parameter
            ).call(block_identifier=block_identifier)

            if require_success and b"" in (data for _, data in result):
                # `b''` values are decoding errors/missing contracts/missing functions
                raise MulticallFailed

            return [
                MulticallResult(success, data if data else None)
                for success, data in result
            ]
        except (ContractLogicError, OverflowError, ValueError):
            raise MulticallFailed

    async def try_aggregate(
        self,
        contract_functions: Sequence[AsyncContractFunction],
        require_success: bool = False,
        block_identifier: BlockIdentifier = "latest",
    ) -> list[MulticallDecodedResult]:
        """
        Calls ``try_aggregate`` on MakerDAO's Multicall contract.

        :param require_success: If ``True``, an exception in any of the functions will stop the execution
        """
        targets_with_data, output_types = self._build_payload(contract_functions)
        results = await self._try_aggregate(
            targets_with_data,
            require_success=require_success,
            block_identifier=block_identifier,
        )
        return [
            MulticallDecodedResult(
                multicall_result.success,
                self._decode_data(output_type, multicall_result.return_data)
                if multicall_result.success
                else multicall_result.return_data,
            )
            for output_type, multicall_result in zip(output_types, results)
        ]



import json
from typing import Iterable
from itertools import islice

from eth_typing import HexStr


def chunks(elements: Iterable, n: int) -> Iterable[list]:
    """
    :param elements: Iterable
    :param n: Number per chunk
    :return: Yield successive n-sized chunks from elements
    """
    it = iter(elements)
    while True:
        chunk = list(islice(it, n))
        if not chunk:
            return
        yield chunk


def link_by_tx_hash(explorer_url: str, tx_hash: HexStr | str):
    return f"{explorer_url}/tx/{tx_hash}"


def to_json(obj) -> str:
    return json.dumps(obj, separators=(',', ':'), ensure_ascii=True)



from eth_typing import ChecksumAddress
from web3.contract.async_contract import AsyncContractFunction
from web3.types import Wei

from ._abi import DISPERSE_ABI
from .contract import Contract

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..chain import Chain


DISPERSE_CONTRACT_ADDRESS = "0xD152f549545093347A162Dce210e7293f1452150"


class Disperse(Contract):
    def __init__(
        self,
        chain: "Chain",
        address: ChecksumAddress | str = None,
        abi=None,
    ):
        address = address or DISPERSE_CONTRACT_ADDRESS
        abi = abi or DISPERSE_ABI
        super().__init__(chain, address, abi)

    def disperse_ether(
            self,
            recipients: list[ChecksumAddress | str],
            values: list[Wei | int],
    ) -> AsyncContractFunction:
        """
        disperseEther(address[] recipients, uint256[] values)
        """
        return self.functions.disperseEther(recipients, values)



from web3 import AsyncHTTPProvider

from .proxy import Proxy


class CustomAsyncHTTPProvider(AsyncHTTPProvider):
    def __init__(
            self,
            *args,
            proxy: Proxy = None,
            **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._proxy = None
        self.set_proxy(proxy)

    def set_proxy(self, proxy: Proxy | None):
        if proxy:
            self._proxy = proxy
            self._request_kwargs["proxies"] = {'https': proxy.as_url, 'http': proxy.as_url}



from typing import TYPE_CHECKING

from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address
from web3 import Web3
from web3.contract.async_contract import (
    AsyncContract,
    AsyncContractFunctions,
    AsyncContractEvents,
)
from web3.types import ABI

if TYPE_CHECKING:
    from ..chain import Chain


class Contract:
    def __init__(self, chain: "Chain", address: ChecksumAddress | str, abi):
        if isinstance(address, str):
            address = to_checksum_address(address)
        self._chain = chain
        self._contract: AsyncContract = self._chain.w3.eth.contract(address, abi=abi)

    @property
    def w3(self) -> Web3:
        return self._chain.w3

    @property
    def chain(self) -> "Chain":
        return self._chain

    @property
    def contract(self) -> AsyncContract:
        return self._contract

    @property
    def address(self) -> ChecksumAddress:
        return self._contract.address

    @property
    def abi(self) -> ABI:
        return self._contract.abi

    @property
    def functions(self) -> AsyncContractFunctions:
        return self._contract.functions

    @property
    def events(self) -> AsyncContractEvents:
        return self._contract.events



from functools import cached_property
from pathlib import Path
from typing import Iterable

from eth_account.account import Account, LocalAccount
from eth_typing import ChecksumAddress, HexStr
from eth_utils import from_wei
from web3.types import TxReceipt, Wei

from . import Chain
from .utils import sign_message, load_lines


class Wallet:
    def __init__(
            self,
            account: LocalAccount,
            *,
            tags: Iterable[str] = None,
    ):
        self.account = account
        self.tags = set(tags) if tags else set()

    def __repr__(self):
        return f"Wallet(address={self.address})"

    def __str__(self) -> str:
        info = f"[{self.address}]"
        if self.tags: info += f" ({', '.join((str(tag) for tag in self.tags))})"
        return info

    @classmethod
    def generate(cls, extra_entropy: str = "", tags: Iterable[str] = None) -> "Wallet":
        account = Account.create(extra_entropy)
        return cls(account, tags=tags)

    @classmethod
    def from_key(cls, private_key: str, tags: Iterable[str] = None) -> "Wallet":
        account = Account.from_key(private_key)
        return cls(account, tags=tags)

    @classmethod
    def from_mnemonic(
            cls,
            mnemonic: str,
            passphrase: str = "",
            tags: Iterable[str] = None,
    ) -> "Wallet":
        account = Account.from_key(mnemonic, passphrase)
        return cls(account, tags=tags)

    @classmethod
    def from_file(cls, filepath: Path | str) -> list["Wallet"]:
        return [cls.from_key(private_key) for private_key in load_lines(filepath)]

    @cached_property
    def private_key(self) -> str:
        return self.account.key.hex()

    @cached_property
    def address(self) -> ChecksumAddress:
        return self.account.address

    @cached_property
    def short_address(self) -> str:
        start = self.account.address[:6]
        end = self.account.address[-1:-4:-1]
        return f"{start}...{end}"

    def sign_message(self, message: str) -> str:
        return sign_message(message, self.account)

    def tx_hash(self, chain: Chain, tx_hash: HexStr | str, value: Wei | int = None) -> str:
        tx_hash_link = chain.get_link_by_tx_hash(tx_hash)
        message = f"{self} {chain} {tx_hash_link}"
        if value is not None:
            message += f"\n\tSent: {from_wei(value, 'ether')} {chain.token.symbol}"
        return message

    def tx_receipt(self, chain: Chain, tx_receipt: TxReceipt | str, value: Wei | int = None) -> str:
        tx_hash = tx_receipt.transactionHash.hex()
        message = self.tx_hash(chain, tx_hash, value)
        tx_fee_wei = tx_receipt.gasUsed * tx_receipt.effectiveGasPrice
        tx_fee = from_wei(tx_fee_wei, "ether")
        message += f"\n\tFee: {tx_fee} {chain.token.symbol}"
        # position = tx_receipt.transactionIndex
        # message += f"\n\tPosition (Transaction index): {position}"
        status = tx_receipt.status
        message += f"\n\tStatus: {status}"
        return message



import asyncio
from dataclasses import dataclass

from eth_account.signers.local import LocalAccount
from eth_typing import BlockNumber, ChecksumAddress, HexStr, Address, Hash32
from hexbytes import HexBytes
from web3 import AsyncWeb3
from web3.contract.async_contract import AsyncContractFunction
from web3.middleware import async_geth_poa_middleware, async_simple_cache_middleware
from web3.types import (
    BlockData,
    BlockIdentifier,
    Nonce,
    TxData,
    TxParams,
    TxReceipt,
    Wei,
)

from .proxy import Proxy
from .provider import CustomAsyncHTTPProvider
from .batch_call import BatchCallManager
from .contract import Contract, Multicall, Disperse, ERC20, ERC721
from .utils import link_by_tx_hash


@dataclass
class NativeToken:
    symbol: str = "ETH"
    decimals: int = 18


class Chain:
    def __init__(
            self,
            rpc: str,
            *,
            name: str = "EVM Chain",
            is_testnet: bool = False,
            use_eip1559: bool = True,
            # Proxy
            proxy: Proxy = None,
            # Native token
            symbol: str = "ETH",
            decimals: int = 18,
            # Explorer
            explorer_url: str = None,
            # Connection settings
            provider_timeout: int = 15,
            # Middlewares
            use_poa_middleware: bool = True,
            use_cache_middleware: bool = True,
            # Multicall
            multicall_v3_contract_address: ChecksumAddress | str = None,
            # Disperse
            disperse_contract_address: ChecksumAddress | str = None,
            # Batch request
            batch_request_size: int = 10,
            batch_request_delay: int = 1,
            # Tx params
            # legacy pricing
            gas_price: Wei | int = None,
            # dynamic fee pricing
            max_fee_per_gas: Wei | int = None,
            max_priority_fee_per_gas: Wei | int = None,
    ):
        self._rpc = rpc
        self.name = name
        self.is_testnet = is_testnet
        self.token = NativeToken(symbol=symbol, decimals=decimals)
        self.explorer_url = explorer_url
        self.use_eip1559 = use_eip1559

        self.w3_provider = CustomAsyncHTTPProvider(self._rpc, proxy=proxy, request_kwargs={"timeout": provider_timeout})
        self.w3 = AsyncWeb3(provider=self.w3_provider)

        if use_poa_middleware:
            self.w3.middleware_onion.inject(async_geth_poa_middleware, layer=0)

        if use_cache_middleware:
            self.w3.middleware_onion.add(async_simple_cache_middleware)

        self.multicall = Multicall(chain=self, address=multicall_v3_contract_address)
        self.disperse = Disperse(chain=self, address=disperse_contract_address)

        self.batch_request = BatchCallManager(
            self, batch_request_size, batch_request_delay)

        self.default_gas_price = gas_price
        self.default_max_fee_per_gas = max_fee_per_gas
        self.default_max_priority_fee_per_gas = max_priority_fee_per_gas

    def __repr__(self):
        return f"Chain(rpc=\"{self.rpc})\""

    def __str__(self):
        return f"<{self.name}>"

    @property
    def rpc(self):
        return self._rpc

    def get_link_by_tx_hash(self, tx_hash: HexBytes | HexStr | str):
        if self.explorer_url is None:
            raise ValueError("Set explorer_url before using this method")

        if isinstance(tx_hash, HexBytes):
            tx_hash = tx_hash.hex()
        return link_by_tx_hash(self.explorer_url, tx_hash)

    ################################################################################
    # Contract creation shortcuts
    ################################################################################

    def contract(self, address, abi) -> Contract:
        return Contract(self, address, abi)

    def erc20(self, address) -> ERC20:
        return ERC20(self, address)

    def erc721(self, address) -> ERC721:
        return ERC721(self, address)

    ################################################################################
    # Shortcuts
    ################################################################################

    async def chain_id(self) -> int:
        return await self.w3.eth.chain_id

    async def client_version(self) -> str:
        return await self.w3.client_version

    async def is_eip1559_supported(self) -> bool:
        """
        :return: True if EIP1559 is supported by the node, False otherwise
        """
        last_block = await self.get_block("latest")
        return True if "baseFeePerGas" in last_block else False

    async def is_contract(self, contract_address: Address | ChecksumAddress | str) -> bool:
        return bool(await self.w3.eth.get_code(contract_address))

    async def get_current_block_number(self) -> BlockNumber:
        return await self.w3.eth.block_number

    async def get_nonce(self, address: Address | ChecksumAddress | str) -> int:
        return await self.w3.eth.get_transaction_count(address)

    async def get_balance(
            self,
            address: Address | ChecksumAddress | str,
            block_identifier: BlockIdentifier = "latest",
    ) -> Wei:
        return await self.w3.eth.get_balance(address, block_identifier)

    async def get_tx(self, tx_hash: Hash32 | HexBytes | HexStr) -> TxData:
        return await self.w3.eth.get_transaction(tx_hash)

    async def get_tx_receipt(self, tx_hash: Hash32 | HexBytes | HexStr) -> TxReceipt:
        return await self.w3.eth.get_transaction_receipt(tx_hash)

    async def wait_for_tx_receipt(
            self,
            tx_hash: Hash32 | HexBytes | HexStr,
            timeout: float = 120,
            poll_latency: float = 0.1
    ) -> TxReceipt:
        """
        raises: TimeExhausted:
                 Raised when a method has not retrieved the desired result within a specified timeout.
                 TransactionNotFound:
                 Raised when a tx hash used to look up a tx in a jsonrpc call cannot be found.
        """
        tx_receipt = await self.w3.eth.wait_for_transaction_receipt(tx_hash, timeout, poll_latency)

        # Add extra sleep to let tx propogate correctly
        await asyncio.sleep(1)
        return tx_receipt

    async def get_block(
            self,
            block_identifier: BlockIdentifier,
            full_transactions: bool = False
    ) -> BlockData:
        return await self.w3.eth.get_block(block_identifier, full_transactions=full_transactions)

    async def check_tx_with_confirmations(
            self,
            tx_hash: Hash32 | HexBytes | HexStr,
            confirmations: int
    ) -> bool:
        tx_receipt = await self.get_tx_receipt(tx_hash)

        if not tx_receipt or tx_receipt["blockNumber"] is None:
            # If `tx_receipt` exists but `blockNumber` is `None`,
            # the transaction is still pending (only for Parity).
            return False
        else:
            block_number = await self.w3.eth.block_number
            confirmations_count = block_number - tx_receipt["blockNumber"]

            return confirmations_count >= confirmations

    async def transfer(
            self,
            account_from: LocalAccount,
            address_to: Address | ChecksumAddress | str,
            value: Wei | int,
            *,
            gas: int = None,
            nonce: Nonce | int = None,
            # legacy pricing
            gas_price: Wei | int = None,
            # dynamic fee pricing
            max_fee_per_gas: Wei | int = None,
            max_priority_fee_per_gas: Wei | int = None,
    ) -> HexStr:
        gas_price = gas_price or self.default_gas_price
        tx_params = self._build_tx_base_params(gas, account_from.address, address_to, nonce, value)
        gas = await self.w3.eth.estimate_gas(tx_params)
        tx_params = self._build_tx_base_params(gas, tx_params=tx_params)
        tx_params = self._build_tx_fee_params(
            gas_price, max_fee_per_gas, max_priority_fee_per_gas, tx_params=tx_params)
        return await self.sign_and_send_tx(account_from, tx_params)

    async def transfer_all(
            self,
            account_from: LocalAccount,
            address_to: Address | ChecksumAddress | str,
            *,
            gas: int = None,
            nonce: Nonce | int = None,
            # legacy pricing
            gas_price: Wei | int = None,
            # dynamic fee pricing
            max_fee_per_gas: Wei | int = None,
            max_priority_fee_per_gas: Wei | int = None,
    ):
        raise NotImplementedError  # TODO Реализовать метод Chain.transfer_all()

    ################################################################################
    # Gas price shortcuts
    ################################################################################

    async def request_gas_price(self) -> Wei:
        return await self.w3.eth.gas_price

    async def request_max_priority_fee(self) -> Wei:
        return await self.w3.eth.max_priority_fee

    ################################################################################
    # Working with transactions
    ################################################################################

    async def _send_tx(self, tx: TxParams) -> HexBytes:
        return await self.w3.eth.send_transaction(tx)

    async def _send_raw_tx(self, raw_tx: bytes | HexStr) -> HexBytes:
        return await self.w3.eth.send_raw_transaction(bytes(raw_tx))

    async def _build_tx_base_params(
            self,
            gas: int = None,
            address_from: Address | ChecksumAddress | str = None,
            address_to: Address | ChecksumAddress | str = None,
            nonce: Nonce | int = None,
            value: Wei | int = None,
            *,
            tx_params: TxParams = None,
    ) -> TxParams:
        if tx_params is None:
            tx_params = dict()
        tx_params = tx_params.copy()

        tx_params["chainId"] = await self.chain_id()

        if gas is not None:
            tx_params["gas"] = gas
        if address_from is not None:
            tx_params["from"] = address_from
        if address_to is not None:
            tx_params["to"] = address_to
        if value is not None:
            tx_params["value"] = value

        if nonce is not None:
            tx_params["nonce"] = nonce
        elif address_from is not None:
            tx_params["nonce"] = await self.get_nonce(address_from)

        return tx_params

    async def _build_tx_fee_params(
            self,
            # legacy pricing
            gas_price: Wei | int = None,
            # dynamic fee pricing
            max_fee_per_gas: Wei | int = None,
            max_priority_fee_per_gas: Wei | int = None,
            *,
            tx_params: TxParams = None,
    ) -> TxParams:
        if tx_params is None:
            tx_params = dict()
        tx_params = tx_params.copy()

        max_fee_per_gas = max_fee_per_gas or self.default_max_fee_per_gas
        max_priority_fee_per_gas = max_priority_fee_per_gas or self.default_max_priority_fee_per_gas

        if (self.is_eip1559_supported and
                ((gas_price is None and self.use_eip1559)
                 or max_fee_per_gas is not None
                 or max_priority_fee_per_gas is not None)):
            tx_params["maxFeePerGas"] = max_fee_per_gas or await self.request_gas_price()
            tx_params["maxPriorityFeePerGas"] = max_priority_fee_per_gas or self.request_max_priority_fee()
        else:
            tx_params["gasPrice"] = gas_price or await self.request_gas_price()

        return tx_params

    async def build_tx(
            self,
            contract_function: AsyncContractFunction,
            gas: int = None,
            address_from: Address | ChecksumAddress | str = None,
            address_to: Address | ChecksumAddress | str = None,
            nonce: Nonce | int = None,
            value: Wei | int = None,
            # legacy pricing
            gas_price: Wei | int = None,
            # dynamic fee pricing
            max_fee_per_gas: Wei | int = None,
            max_priority_fee_per_gas: Wei | int = None,
    ) -> TxParams:
        gas_price = gas_price or self.default_gas_price
        tx_params = await self._build_tx_base_params(gas, address_from, address_to, nonce, value)
        gas = await contract_function.estimate_gas(tx_params)
        tx_params = await self._build_tx_base_params(gas, tx_params=tx_params)
        tx_params = await self._build_tx_fee_params(
            gas_price, max_fee_per_gas, max_priority_fee_per_gas, tx_params=tx_params)
        return await contract_function.build_transaction(tx_params)

    async def sign_and_send_tx(self, account: LocalAccount, tx: TxParams) -> HexStr:
        signed_tx = account.sign_transaction(tx)
        tx_hash = await self._send_raw_tx(signed_tx.rawTransaction)
        return HexStr(tx_hash.hex())

    async def execute_fn(
            self,
            account: LocalAccount,
            fn: AsyncContractFunction,
            *,
            gas: int = None,
            nonce: Nonce | int = None,
            value: Wei | int = None,
            # legacy pricing
            gas_price: Wei | int = None,
            # dynamic fee pricing
            max_fee_per_gas: Wei | int = None,
            max_priority_fee_per_gas: Wei | int = None,
    ) -> HexStr:
        tx = await self.build_tx(
            fn,
            address_from=account.address,
            gas=gas,
            nonce=nonce,
            value=value,
            gas_price=gas_price,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
        )
        tx_hash = await self.sign_and_send_tx(account, tx)
        return tx_hash



from functools import cached_property
from typing import TYPE_CHECKING, Iterable

from eth_typing import ChecksumAddress, BlockIdentifier
from web3.types import Wei

from ._abi import ERC20_ABI
from .contract import Contract

if TYPE_CHECKING:
    from ..chain import Chain


class ERC20(Contract):
    def __init__(
            self,
            chain: "Chain",
            address: ChecksumAddress | str,
            abi=None,
    ):
        abi = abi or ERC20_ABI
        super().__init__(chain, address, abi)

    @cached_property
    async def name(self) -> str:
        return await self.functions.name().call()

    @cached_property
    async def symbol(self) -> str:
        return await self.functions.symbol().call()

    @cached_property
    async def decimals(self) -> int:
        return await self.functions.decimals().call()

    async def get_balance(
            self,
            address: ChecksumAddress,
            block_identifier: BlockIdentifier = "latest",
    ) -> Wei:
        balance = await self.functions.balanceOf(address).call(block_identifier=block_identifier)
        return balance

    async def get_balances(
            self,
            addresses: Iterable[ChecksumAddress],
            block_identifier: BlockIdentifier = "latest",
            **kwargs,
    ) -> Iterable[dict[str: ChecksumAddress | str, str: Wei]]:
        if not addresses:
            return

        balances = await self.chain.batch_request.contract_request(
            [self.functions.balanceOf(address) for address in addresses],
            block_identifier,
            **kwargs,
        )
        for address, balance_data in zip(addresses, balances):
            yield {"address": address, "balance":  balance_data["result"]}



from .eth import (
    sign_message,
    decode_string_or_bytes32,
    to_checksum_addresses,
    addresses_from_file,
    estimate_data_gas,
    hex_block_identifier,
)
from .file import (
    copy_file,
    load_lines,
    load_json,
    load_toml,
    write_lines,
    write_json,
)
from .other import (
    chunks,
    link_by_tx_hash,
    to_json,
)


__all__ = [
    "sign_message",
    "decode_string_or_bytes32",
    "to_checksum_addresses",
    "addresses_from_file",
    "estimate_data_gas",
    "hex_block_identifier",
    "copy_file",
    "load_lines",
    "load_json",
    "load_toml",
    "write_lines",
    "write_json",
    "chunks",
    "link_by_tx_hash",
    "to_json",
]



import re
from pathlib import Path
from typing import Iterable

from .utils import load_lines


# PROTOCOLS = ["http", "https", "SOP2", "SOP3", "socks4", "socks5"]
PROXY_FORMATS_REGEXP = [
    re.compile(r'^(?:(?P<type>.+)://)?(?P<login>[^:]+):(?P<password>[^@|:]+)[@|:](?P<ip>[^:]+):(?P<port>\d+)$'),
    re.compile(r'^(?:(?P<type>.+)://)?(?P<ip>[^:]+):(?P<port>\d+)[@|:](?P<login>[^:]+):(?P<password>[^:]+)$'),
    re.compile(r'^(?:(?P<type>.+)://)?(?P<ip>[^:]+):(?P<port>\d+)$'),
]


class Proxy:
    """
    A class to represent a proxy connection.

    Attributes
    ----------
    ip : str
        IP address of the proxy server
    port : int
        Port to use for the proxy server
    protocol : str, optional
        Protocol to use for the proxy connection, by default "http"
    login : str, optional
        Login for the proxy server, by default None
    password : str, optional
        Password for the proxy server, by default None
    """

    def __init__(
            self,
            ip: str,
            port: int,
            *,
            tags: Iterable[str] = None,
            protocol: str = None,
            login: str = None,
            password: str = None,
    ):
        self.protocol = protocol or "http"
        self.ip = ip
        self.port = port
        self.login = login
        self.password = password
        self.tags = set(tags) if tags else set()

    @classmethod
    def from_str(cls, proxy: str, *, tags: Iterable[str] = None) -> "Proxy":
        """
        Class method to create a Proxy object from a string.

        The string can be in one of the following formats:
        - '(type://)?ip:port[:@|]login:password'
        - '(type://)?login:password[:@|]ip:port'

        Parameters
        ----------
            proxy : str
                Proxy connection string in one of the above formats.

        Returns
        -------
            Proxy
                A Proxy object.

        Examples
        --------
        >>> proxy = Proxy.from_str('192.168.1.1:8080')
        >>> print(proxy)
        http://192.168.1.1:8080

        >>> proxy = Proxy.from_str('https://user:pass@192.168.1.1:8080')
        >>> print(proxy)
        https://192.168.1.1:8080@user:pass
        """
        for pattern in PROXY_FORMATS_REGEXP:
            match = pattern.match(proxy)
            if match:
                return cls(
                    ip=match.group('ip'),
                    port=int(match.group('port')),
                    protocol=match.group('type'),
                    login=match.group('login'),
                    password=match.group('password'),
                    tags=tags,
                )

        raise ValueError(f'Unsupported proxy format: {proxy}')

    @classmethod
    def from_file(cls, filepath: Path | str) -> set["Proxy"]:
        return {cls.from_str(proxy) for proxy in load_lines(filepath)}

    @property
    def as_url(self) -> str:
        return (f"{self.protocol}://"
                + (f"{self.login}:{self.password}@" if self.login and self.password else "")
                + f"{self.ip}:{self.port}")

    def __repr__(self):
        return f"Proxy(ip={self.ip}, port={self.port})"

    def __str__(self) -> str:
        info = f"[{self.ip:>15}:{str(self.port):<5}]"
        if self.tags: info += f" ({', '.join((str(tag) for tag in self.tags))})"
        return info

    def __hash__(self):
        return hash((self.ip, self.port, self.protocol, self.login, self.password))

    def __eq__(self, other):
        if isinstance(other, Proxy):
            return (
                self.ip == other.ip
                and self.port == other.port
                and self.protocol == other.protocol
                and self.login == other.login
                and self.password == other.password
            )
        return False



from functools import cached_property
from typing import TYPE_CHECKING, Iterable

from eth_typing import ChecksumAddress, BlockIdentifier
from eth_utils import to_checksum_address

from ._abi import ERC721_ABI
from .contract import Contract

if TYPE_CHECKING:
    from ..chain import Chain


class ERC721(Contract):
    def __init__(
            self,
            chain: "Chain",
            address: ChecksumAddress | str,
            abi=None,
    ):
        abi = abi or ERC721_ABI
        super().__init__(chain, address, abi)

    @cached_property
    def name(self) -> str:
        return self.functions.name().call()

    @cached_property
    def symbol(self) -> str:
        return self.functions.symbol().call()

    def get_balance(
            self,
            address: ChecksumAddress,
            block_identifier: BlockIdentifier = "latest",
    ) -> int:
        return self.functions.balanceOf(address).call(block_identifier=block_identifier)

    async def get_balances(
            self,
            addresses: Iterable[ChecksumAddress],
            block_identifier: BlockIdentifier = "latest",
            **kwargs,
    ) -> Iterable[dict[str: ChecksumAddress, str: int]]:
        if not addresses:
            return

        balances = await self.chain.batch_request.contract_request(
            [self.functions.balanceOf(address) for address in addresses],
            block_identifier,
            **kwargs,
        )
        for address, balance_data in zip(addresses, balances):
            yield {"address": address, "balance":  balance_data["result"]}

    async def get_owners(
            self,
            token_ids: Iterable[int],
            block_identifier: BlockIdentifier = "latest",
            **kwargs,
    ) -> Iterable[dict[str: int, str: ChecksumAddress]]:
        if not token_ids:
            return

        owners = await self.chain.batch_request.contract_request(
            [self.functions.ownerOf(token_id) for token_id in token_ids],
            block_identifier,
            **kwargs,
        )
        for token_id, owner_data in zip(token_ids, owners):
            yield {"token_id": token_id, "owner": to_checksum_address(owner_data["result"])}

    async def get_token_uris(
            self,
            token_ids: Iterable[int],
            block_identifier: BlockIdentifier = "latest",
            **kwargs,
    ) -> Iterable[dict[str: int, str: str]]:
        if not token_ids:
            return

        token_uris = await self.chain.batch_request.contract_request(
            [self.functions.tokenURI(token_id) for token_id in token_ids],
            block_identifier=block_identifier,
            **kwargs,
        )
        for token_id, uri_data in zip(token_ids, token_uris):
            yield {"token_id": token_id, "uri": uri_data["result"]}
